// Original dramarduino project by iss (forum.defence-force.org)
// Corrected and speed optimized by martin@8bity.cz
// Adapted for Arduino UNO and Arduino Leonardo
// Last updated 30 Mar 2019 18:19

// #include <SoftwareSerial.h>

#define DI          A1  // PC1 PF6
#define DO           8  // PB0 PB4
#define CAS          9  // PB1 PB5
#define RAS         A3  // PC3 PF4
#define WE          A2  // PC2 PF5

#define M_TYPE      10  // PB2 PB6
#define R_LED       11  // PB3 PB7
#define G_LED       12  // PB4 PD6

#define RXD          0  // PD0 PD2
#define TXD          1  // PD1 PD3

#define BUS_SIZE     9

#define REFRESH_PAUSE 50  // pause without refresh between March cycles in ms

volatile int bus_size;

// SoftwareSerial USB(RXD, TXD);

void setBus(unsigned int addr) {
#ifdef ARDUINO_AVR_UNO
  // set address bit x - pin = uno port / leonardo port
  // set address bit 0 - A4 = PC4 / PF1
  if( addr & 0b000000001)
    PORTC |= B00010000;
  else
    PORTC &= B11101111;

  // set address bit 1 - D2 = PD2 / PD1
  if(addr & 0b000000010)
    PORTD |= B00000100;
  else
    PORTD &= B11111011;

  // set address bit 2 - A5 = PC5 / PF0
  if(addr & 0b000000100)
    PORTC |= B00100000;
  else
    PORTC &= B11011111;
  
  // set address bit 3 - D6 = PD6 / PD7
  if(addr & 0b000001000)
    PORTD |= B01000000;
  else
    PORTD &= B10111111;

  // set address bit 4 - D5 = PD5 / PC6
  if(addr & 0b000010000)
    PORTD |= B00100000;
  else
    PORTD &= B11011111;

  // set address bit 5 - D4 = PD4 / PD4
  if(addr & 0b000100000)
    PORTD |= B00010000;
  else
    PORTD &= B11101111;

  // set address bit 6 - D7 = PD7 / PE6
  if(addr & 0b001000000)
    PORTD |= B10000000;
  else
    PORTD &= B01111111;

  // set address bit 7 - D3 = PD3 / PD0
  if(addr & 0b010000000)
    PORTD |= B00001000;
  else
    PORTD &= B11110111;

  // set address bit 8 - A0 = PC0 / PF7
  if(addr & 0b100000000)
    PORTC |= B00000001;
  else
    PORTC &= B11111110;
#endif

#ifdef ARDUINO_AVR_LEONARDO
  // set address bit x - pin = uno port / leonardo port
  // set address bit 0 - A4 = PC4 / PF1
  if( addr & 0b000000001)
    PORTF |= B00000010;
  else
    PORTF &= B11111101;

  // set address bit 1 - D2 = PD2 / PD1
  if(addr & 0b000000010)
    PORTD |= B00000010;
  else
    PORTD &= B11111101;

  // set address bit 2 - A5 = PC5 / PF0
  if(addr & 0b000000100)
    PORTF |= B00000001;
  else
    PORTF &= B11111110;
  
  // set address bit 3 - D6 = PD6 / PD7
  if(addr & 0b000001000)
    PORTD |= B10000000;
  else
    PORTD &= B01111111;

  // set address bit 4 - D5 = PD5 / PC6
  if(addr & 0b000010000)
    PORTC |= B01000000;
  else
    PORTC &= B10111111;

  // set address bit 5 - D4 = PD4 / PD4
  if(addr & 0b000100000)
    PORTD |= B00010000;
  else
    PORTD &= B11101111;

  // set address bit 6 - D7 = PD7 / PE6
  if(addr & 0b001000000)
    PORTE |= B01000000;
  else
    PORTE &= B10111111;

  // set address bit 7 - D3 = PD3 / PD0
  if(addr & 0b010000000)
    PORTD |= B00000001;
  else
    PORTD &= B11111110;

  // set address bit 8 - A0 = PC0 / PF7
  if(addr & 0b100000000)
    PORTF |= B10000000;
  else
    PORTF &= B01111111;
#endif
}

void writeAddress(unsigned int r, unsigned int c, int v) {
  /* row */
  setBus(r);
  // digitalWrite(RAS, LOW); PC3 / PF4
  #ifdef ARDUINO_AVR_UNO
    PORTC &= B11110111;
  #elif ARDUINO_AVR_LEONARDO
    PORTF &= B11101111;
  #endif

  /* rw */
  // digitalWrite(WE, LOW); PC2 / PF5
  #ifdef ARDUINO_AVR_UNO
    PORTC &= B11111011;
  #elif ARDUINO_AVR_LEONARDO
    PORTF &= B11011111;
  #endif

  /* val */
  // digitalWrite(DI, (v & 1)? HIGH : LOW); PC1 / PF6
  #ifdef ARDUINO_AVR_UNO
    if(v & 1)
      PORTC |= B00000010;
    else
      PORTC &= B11111101;
  #elif ARDUINO_AVR_LEONARDO
    if(v & 1)
      PORTF |= B01000000;
    else
      PORTF &= B10111111;
  #endif

  /* col */
  setBus(c);
  // digitalWrite(CAS, LOW); PB1 / PB5
  #ifdef ARDUINO_AVR_UNO
    PORTB &= B11111101;
    PORTB &= B11111101; // wait for sure ;-)
  #elif ARDUINO_AVR_LEONARDO
    PORTB &= B11011111;
    PORTB &= B11011111; // wait for sure ;-)
  #endif

  // digitalWrite(WE, HIGH); PC2 / PF5
  #ifdef ARDUINO_AVR_UNO
    PORTC |= B00000100;
  #elif ARDUINO_AVR_LEONARDO
    PORTF |= B00100000;
  #endif
  // digitalWrite(CAS, HIGH); PB1 / PB5
  #ifdef ARDUINO_AVR_UNO
    PORTB |= B00000010;
  #elif ARDUINO_AVR_LEONARDO
    PORTB |= B00100000;
  #endif
  // digitalWrite(RAS, HIGH);  PC3 / PF4
  #ifdef ARDUINO_AVR_UNO
    PORTC |= B00001000;
  #elif ARDUINO_AVR_LEONARDO
    PORTF |= B00010000;
  #endif
}

int readAddress(unsigned int r, unsigned int c) {
  int ret = 0;

  /* row */
  setBus(r);
  // digitalWrite(RAS, LOW); PC3 / PF4
  #ifdef ARDUINO_AVR_UNO
    PORTC &= B11110111;
  #elif ARDUINO_AVR_LEONARDO
    PORTF &= B11101111;
  #endif

  /* col */
  setBus(c);
  // digitalWrite(CAS, LOW); PB1 / PB5
  #ifdef ARDUINO_AVR_UNO
    PORTB &= B11111101;
    PORTB &= B11111101; // wait for sure ;-)
  #elif ARDUINO_AVR_LEONARDO
    PORTB &= B11011111;
    PORTB &= B11011111; // wait for sure ;-)
  #endif

  /* get current value */
  // ret = digitalRead(DO); PB0 / PB4
  #ifdef ARDUINO_AVR_UNO
    ret = PINB & B00000001;
  #elif ARDUINO_AVR_LEONARDO
    ret = (PINB & B00010000)>>4;
  #else
    ret = 0;
  #endif

  // digitalWrite(CAS, HIGH); PB1 / PB5
  #ifdef ARDUINO_AVR_UNO
    PORTB |= B00000010;
  #elif ARDUINO_AVR_LEONARDO
    PORTB |= B00100000;
  #endif
  // digitalWrite(RAS, HIGH);  PC3 / PF4
  #ifdef ARDUINO_AVR_UNO
    PORTC |= B00001000;
  #elif ARDUINO_AVR_LEONARDO
    PORTF |= B00010000;
  #endif

  return ret;
}

void error(int r, int c)
{
  setBus(0); 
  unsigned long a = ((unsigned long)c << bus_size) + r;
  digitalWrite(R_LED, LOW);
  digitalWrite(G_LED, HIGH);
  interrupts();
  Serial.print(" FAILED $");
  Serial.println(a, HEX);
  Serial.flush();
  while (1)
    ;
}

void ok(void)
{
  setBus(0); 
  digitalWrite(R_LED, HIGH);
  digitalWrite(G_LED, LOW);
  interrupts();
  Serial.println(" OK!");
  Serial.flush();
  while (1)
    ;
}

void blink(void)
{
  interrupts();
  digitalWrite(G_LED, LOW);
  digitalWrite(R_LED, LOW);
  // refresh off between March phases
  delay(REFRESH_PAUSE);
  digitalWrite(R_LED, HIGH);
  digitalWrite(G_LED, HIGH);
  noInterrupts();
}

void green(int v) {
  // digitalWrite(G_LED, v); PB4 / PD6
  #ifdef ARDUINO_AVR_UNO
    if(v & 1)
      PORTB |= B00010000;
    else
      PORTB &= B11101111;
  #elif ARDUINO_AVR_LEONARDO
    if(v & 1)
      PORTD |= B01000000;
    else
      PORTD &= B10111111;
  #endif
}

void fill(int v) {
  int r, c, g = 0;
  v %= 1;
  for (c = 0; c < (1<<bus_size); c++) {
    green(g? HIGH : LOW);
    for (r = 0; r < (1<<bus_size); r++) {
      writeAddress(r, c, v);
      if (v != readAddress(r, c))
        error(r, c);
    }
    g ^= 1;
  }
  blink();
}

void fillx(int v) {
  int r, c, g = 0;
  v %= 1;
  for (c = 0; c < (1<<bus_size); c++) {
    green(g? HIGH : LOW);
    for (r = 0; r < (1<<bus_size); r++) {
      writeAddress(r, c, v);
      if (v != readAddress(r, c))
        error(r, c);
      v ^= 1;
    }
    g ^= 1;
  }
  blink();
}

void test_MarchCminus() {
  int c, r, memsize;
  memsize = (1<<bus_size);
  // -> w0
  for(c = 0; c < memsize; c++)
    for(r = 0; r < memsize; r++)
      writeAddress(r, c, 0);
  blink();
  // <- r0w1
  for(c = memsize-1; c >= 0; c--)
    for(r = memsize-1; r >= 0; r--) {
      if (readAddress(r, c) != 0)
        error(r, c);
      writeAddress(r, c, 1);
    }
  blink();
  // <- r1w0
  for(c = memsize-1; c >= 0; c--)
    for(r = memsize-1; r >= 0; r--) {
      if (readAddress(r, c) != 1)
        error(r, c);
      writeAddress(r, c, 0);
    }
  blink();
  // -> r0w1
  for(c = 0; c < memsize; c++)
    for(r = 0; r < memsize; r++) {
      if (readAddress(r, c) != 0)
        error(r, c);
      writeAddress(r, c, 1);
    }
  blink();
  // -> r1w0
  for(c = 0; c < memsize; c++)
    for(r = 0; r < memsize; r++) {
      if (readAddress(r, c) != 1)
        error(r, c);
      writeAddress(r, c, 0);
    }
  blink();
  // -> r0
  for(c = 0; c < memsize; c++)
    for(r = 0; r < memsize; r++)
      if (readAddress(r, c) != 0)
        error(r, c);
}
  
void test_MarchY() {
  int c, r, memsize;
  memsize = (1<<bus_size);
  // -> w0
  for(c = 0; c < memsize; c++)
    for(r = 0; r < memsize; r++)
      writeAddress(r, c, 0);
  blink();
  // <- r0w1r1
  for(c = memsize-1; c >= 0; c--)
    for(r = memsize-1; r >= 0; r--) {
      if (readAddress(r, c) != 0)
        error(r, c);
      writeAddress(r, c, 1);
      if (readAddress(r, c) != 1)
        error(r, c);
    }
  blink();
  // -> r1w0r0
  for(c = 0; c < memsize; c++)
    for(r = 0; r < memsize; r++) {
      if (readAddress(r, c) != 1)
        error(r, c);
      writeAddress(r, c, 0);
      if (readAddress(r, c) != 0)
        error(r, c);
    }
  blink();
  // -> r0
  for(c = 0; c < memsize; c++)
    for(r = 0; r < memsize; r++)
      if (readAddress(r, c) != 0)
        error(r, c);
}

void device_Initialization() {
  int i;
  // some devices need 200us wait after start (it took possibly much more time to get here, but just for sure)
  for( i = 0; i < 635; i++)
    __asm__("nop\n\t");
  // some devices need to initialize dynamic circuitry by RAS address strobes, will make CAS too
  for( i = 0; i < 10; i++)
    writeAddress(0, 0, 0);  
}
void test_DataBus() {
  writeAddress(0, 0, 0);
  if(readAddress(0, 0) != 0)
    error(0, 0);
  writeAddress(0, 0, 1);
  if(readAddress(0, 0) != 1)
    error(1, 0);
  writeAddress(0, 0, 0);
  if(readAddress(0, 0) != 0)
    error(2, 0);
}

void test_AddressBus() {
  int r, c, rr, cc, ret, memsize;
  memsize = (1<<bus_size);
  // check all adress lines, it should be OK to check just rows, but checking cols will check DRAM internal address cicuitry too
  for(c = 1; c < memsize; c<<=1)
    for(r = 1; r < memsize; r<<=1) {
      writeAddress(r, c, 1);
      for(cc = 1; cc < memsize; cc<<=1)
        for(rr = 1; rr < memsize; rr<<=1) {
          if(c!=cc || r!=rr) {
            writeAddress(rr, cc, 0);
          }
        }
      writeAddress(0, 0, 0);
      for(cc = 1; cc < memsize; cc<<=1)
        for(rr = 1; rr < memsize; rr<<=1) {
          ret = readAddress(rr, cc);
          if(c==cc && r==rr) {
            if(ret!=1)
              error(rr, cc);
          } else {
            if(ret!=0)
              error(rr, cc);
          }
        }
    }
}

void setup() {
  int i;

  Serial.begin(9600);
  while (!Serial)
    ; /* wait */
    
  Serial.println();
  Serial.print("DRAM tester+ ");

  // set address line pins as output
  pinMode(A4, OUTPUT);
  pinMode(2, OUTPUT);
  pinMode(A5, OUTPUT);
  pinMode(6, OUTPUT);
  pinMode(5, OUTPUT);
  pinMode(4, OUTPUT);
  pinMode(7, OUTPUT);
  pinMode(3, OUTPUT);
  pinMode(A0, OUTPUT);
  // all address pins to 0
  setBus(0);
  
  pinMode(CAS, OUTPUT);
  pinMode(RAS, OUTPUT);
  pinMode(WE, OUTPUT);
  pinMode(DI, OUTPUT);

  pinMode(R_LED, OUTPUT);
  pinMode(G_LED, OUTPUT);

  pinMode(M_TYPE, INPUT);
  pinMode(DO, INPUT);

  digitalWrite(WE, HIGH);
  digitalWrite(RAS, HIGH);
  digitalWrite(CAS, HIGH);

  digitalWrite(R_LED, HIGH);
  digitalWrite(G_LED, HIGH);

  if (digitalRead(M_TYPE)) {
    /* jumper not set - 41256 */
    bus_size = BUS_SIZE;
    Serial.print("256Kx1 ");
  } else {
    /* jumper set - 4164 */
    bus_size = BUS_SIZE - 1;
    Serial.print("64Kx1 ");
  }
  Serial.flush();

  noInterrupts();
}

void loop() {
  
  interrupts();
  #ifdef ARDUINO_AVR_UNO
    Serial.println("Arduino UNO");
  #elif ARDUINO_AVR_LEONARDO
    Serial.println("Arduino Leonardo");
  #else
    Serial.println("Unknown board - does not work");
    error(-1,0);
  #endif

  device_Initialization();

  interrupts(); Serial.print("DataBus "); Serial.flush(); noInterrupts(); test_DataBus();
  interrupts(); Serial.print("AddressBus "); Serial.flush(); noInterrupts(); test_AddressBus();
 
  interrupts(); Serial.print("fillx0 "); Serial.flush(); noInterrupts(); fillx(0);
  interrupts(); Serial.print("fillx1 "); Serial.flush(); noInterrupts(); fillx(1);
  interrupts(); Serial.print("fill0 "); Serial.flush(); noInterrupts(); fill(0);
  interrupts(); Serial.print("fill1 "); Serial.flush(); noInterrupts(); fill(1);

  interrupts(); Serial.print("MarchC- "); Serial.flush(); noInterrupts(); test_MarchCminus();
  interrupts(); Serial.print("MarchY"); Serial.flush(); noInterrupts(); test_MarchY();
  
  ok();
}
